//Cluster Demo on AOS with network support
//Approximates Pi

#include <mega32.h>
#include <delay.h>
#include <stdio.h>

#include "aos.h"  
#include "user_tasks.h"
#include "aos_core.c" 
#include "aos_task.c"                   
#include "aos_semaphore.c"	//omit if no semaphores or mail used
//#include "aos_mbox.c"		//omit if no mailboxes     		
//#include "aos_uart.c"		//omit if you don't need UART
#include "aos_network.c"

#define PID         (UBYTE)0
#define NUMPROCS    (UBYTE)4

void task1(void)
{
    //For status LEDs
    PORTA = 0xff;
    DDRA = 0xff;
    
    //USART Init  
    UCSRB = 0x18;
    UBRRL = 103;
        
	  
    net_init();
    PORTA = 0x00;
    delay_ms(100);
    
    //program
    {
        if (PID == (UBYTE)0)                    //Node 0 (MASTER)
        {
            float PI = 3.141592653589793238462643;
            float mypi, pi, h, sum, x;
            int n = 20, i, j, k;                //n=number of approximation intervals
            UBYTE *c, buf[sizeof(float)*3];
            
            printf("Approximation to Pi=%f\r\n", PI);
            //send n to all workers
            c = (UBYTE*)(&n);                   //get a byte pointer to the int
            PORTA = 0xff;
            putsf("Sending to P1\r");
            net_send(1, 0, sizeof(int), c);
            putsf("Sending to P2\r");
            net_send(2, 0, sizeof(int), c);
            putsf("Sending to P3\r");
            net_send(3, 0, sizeof(int), c);
            PORTA = 0x0f;

            //do the partial sum to the approximation
            h = 1.0 / (float)n;
            sum = 0.0;
            for (i = PID + 1; i <= n; i += NUMPROCS)
            {
                x = h * ((double)i - 0.5);
                sum += (4.0 / (1.0 + x*x));
            }
            mypi = h * sum;
            printf("my partial sum is %f\r\n", mypi);

            //wait for workers to submit partial results
            PORTA = 0xf0;
            for (i = 0; i < sizeof(float) * 3; i++)         //should receive 3 floats, byte by byte
            {
                receive_wait();
                putsf("Received a byte\r");
            }
            PORTA = 0xaa;

            //retrieve the partial results from the buffers
            putsf("Pulling from buffers\r");
            i = j = k = 0;
            while ((i + j + k) != (sizeof(float) * 3))      //should receive 3 floats
            {
                i += net_recv(1, 255, sizeof(float) - i, buf + i);                                      //addr 1 starts at buf
                j += net_recv(2, 255, sizeof(float) - j, buf + sizeof(float) + j);                      //addr 2 offset by 1 float
                k += net_recv(3, 255, sizeof(float) - k, buf + sizeof(float) + sizeof(float) + k);      //addr 3 offset by 2 floats
            }
            
            //sum up results
            pi = mypi + *((float*)buf) + *((float*)(buf + sizeof(float))) + *((float*)(buf + sizeof(float) + sizeof(float)));
            printf("pi is approximately %f, Error is %f\r\n", pi, pi - PI);
        }
        else                                    //Nodes 1-3 (WORKERS)
        {
            float mypi, h, sum, x;
            int n, i;
            UBYTE *c;
            
            putsf("Waiting for n\r");
            //receive n value from master
            for (i = 0; i < sizeof(int); i++)                   //should receive 1 int, byte by byte
            {
                receive_wait();
                putsf("Got a byte\r");
            }

            i = 0;
            c = (UBYTE *)(&n);                                  //byte pointer to the int
            putsf("Pulling from buffer\r");
            while (i != sizeof(int))
                i += net_recv(0, 255, sizeof(int) - i, c);      //retrieve n from the receive buffer

            //do the partial sum approximation
            h = 1.0 / (float)n;
            sum = 0.0;
            for (i = PID + 1; i <= n; i += NUMPROCS)
            {
                x = h * ((double)i - 0.5);
                sum += (4.0 / (1.0 + x*x));
            }
            mypi = h * sum;

            //send the partial result
            c = (UBYTE *)(&mypi);                               //byte pointer to the float
            putsf("Sending the result\r");
            net_send(0, 0, sizeof(float), c);
        }
    }
    
    //done, never end
	while (1)
	{
	}
}